import abc
import functools as ft
import json

from pjrpc import common
from pjrpc.common import generators, exceptions, v20


class Batch:
    """
    Batch wrapper. Implements some methods to wrap multiple JSON-RPC requests into a single batch request.

    :param client: JSON-RPC client instance
    :param strict: if ``True`` checks that all requests have theirs corresponding responses
    """

    class Proxy:
        """
        Proxy object. Provides syntactic sugar to make method calls using dot notation.

        :param batch: batch wrapper
        """

        def __init__(self, batch):
            self._batch = batch

        def __getattr__(self, attr):
            def wrapped(*args, **kwargs):
                self._batch.add(attr, *args, **kwargs)
                return self

            return wrapped

        def __call__(self):
            """
            Makes an RPC call.
            """

            return self.call()

        def call(self):
            """
            Makes an RPC call.
            """

            return self._batch.call()

    @property
    def proxy(self):
        """
        Batch request proxy object.
        """

        return Batch.Proxy(self)

    def __init__(self, client, strict=True):
        self._client = client
        self._strict = strict
        self._id_gen = client.id_gen()
        self._requests = client.batch_request_class()

    def __getitem__(self, requests):
        """
        Adds requests to the batch and makes a request.

        :param requests: requests to be added to the batch
        :returns: request results as a tuple
        """

        self._requests.extend([
            self._client.request_class(
                method=method,
                params=params,
                id=next(self._id_gen)
            ) for method, *params in requests
        ])
        return self.call()

    def __call__(self, method, *args, **kwargs):
        """
        Adds the method call to the batch.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: self
        """

        return self.add(method, *args, **kwargs)

    def add(self, method, *args, **kwargs):
        """
        Adds the method call to the batch.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: self
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        self._requests.append(self._client.request_class(method, args or kwargs, id=next(self._id_gen)))
        return self

    def call(self):
        """
        Makes a JSON-RPC request.

        :returns: request results as a tuple
        """

        request_text = self._client.json_dumper(self._requests, cls=self._client.json_encoder)
        response_text = self._client._request(request_text)

        response = self._client.batch_response_class.from_json(
            self._client.json_loader(response_text, cls=self._client.json_decoder)
        )

        self._relate(self._requests, response)

        return response.result

    def send(self):
        """
        Sends a JSON-RPC request.

        :returns: batch response instance
        """

        return self._client.send(self._requests)

    def _relate(self, batch_request, batch_response):
        """
        Sets requests `related` field. if `strict` flag is ``True``
        checks that all requests have theirs corresponding responses

        :param batch_request: batch request
        :param batch_response: batch response
        """

        response_map = {response.id: response for response in batch_response if response.id is not None}

        for request in batch_request:
            if request.id is not None:
                related = response_map.pop(request.id, None)
                if related is None and self._strict:
                    raise exceptions.IdentityError(f"response '{request.id}' not found")
                else:
                    request.related = related

        if response_map:
            raise exceptions.IdentityError(f"unexpected response found: {response_map.keys()}")


class AsyncBatch(Batch):
    """
    Asynchronous batch wrapper. Used to make asynchronous JSON-RPC batch requests.
    """

    async def call(self):
        """
        Makes a JSON-RPC request.

        :returns: request results as a tuple
        """

        request_text = self._client.json_dumper(self._requests, cls=self._client.json_encoder)
        response_text = await self._client._request(request_text)

        response = self._client.batch_response_class.from_json(
            self._client.json_loader(response_text, cls=self._client.json_decoder)
        )

        self._relate(self._requests, response)

        return response.result


class AbstractClient(abc.ABC):
    """
    Abstract JSON-RPC client.

    :param request_class: request class
    :param response_class: response class
    :param batch_request_class: batch request class
    :param batch_response_class: batch response class
    :param id_gen: identifier generator
    :param json_loader: json loader
    :param json_dumper: json dumper
    :param json_encoder: json encoder
    :param json_decoder: json decoder
    :param error_cls: JSON-RPC error base class
    :param strict: if ``True`` checks that a request and a response identifiers match
    """

    class Proxy:
        """
        Proxy object. Provides syntactic sugar to make method call using dot notation.

        :param client: JSON-RPC client instance
        """

        def __init__(self, client):
            self._client = client

        def __getattr__(self, attr):
            return ft.partial(self._client.call, attr)

    @property
    def proxy(self):
        """
        Clint proxy object.
        """

        return AbstractClient.Proxy(self)

    @property
    def batch(self):
        """
        Client batch wrapper.
        """

        return Batch(self)

    @abc.abstractmethod
    def _request(self, data, **kwargs):
        """
        Makes a JSON-RPC request.

        :param data: request text representation
        :returns: response text representation
        """

    def __init__(
        self,
        request_class=v20.Request,
        response_class=v20.Response,
        batch_request_class=v20.BatchRequest,
        batch_response_class=v20.BatchResponse,
        id_gen=generators.sequential,
        json_loader=json.loads,
        json_dumper=json.dumps,
        json_encoder=common.JSONEncoder,
        json_decoder=None,
        error_cls=exceptions.JsonRpcError,
        strict=True,
        request_args=None
    ):
        self.request_class = request_class
        self.response_class = response_class
        self.batch_request_class = batch_request_class
        self.batch_response_class = batch_response_class
        self.json_loader = json_loader
        self.json_dumper = json_dumper
        self.json_encoder = json_encoder
        self.json_decoder = json_decoder
        self.id_gen = id_gen
        self.error_cls = error_cls
        self.strict = strict
        self.request_args = request_args or {}

    def __call__(self, method, *args, **kwargs):
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        return self.call(method, *args, **kwargs)

    def send(self, request, **kwargs):
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :returns: response instance
        """

        kwargs = {**self.request_args, **kwargs}

        request_text = self.json_dumper(request, cls=self.json_encoder)
        response_text = self._request(request_text, **kwargs)

        return self.response_class.from_json(
            self.json_loader(response_text, cls=self.json_decoder), error_cls=self.error_cls
        )

    def notify(self, method, *args, **kwargs):
        """
        Makes a notification request

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=None,
            method=method,
            params=args or kwargs,
        )
        request_text = self.json_dumper(request, cls=self.json_encoder)
        response_text = self._request(request_text, **self.request_args)
        if self.strict and response_text:
            raise exceptions.BaseError("unexpected response")

    def call(self, method, *args, **kwargs):
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=next(self.id_gen()),
            method=method,
            params=args or kwargs,
        )
        response = self.send(request)
        self._relate(request, response)

        return response.result

    @classmethod
    def _relate(cls, request, response):
        """
        Checks the the request and the response identifiers match.

        :param request: request
        :param response: response
        """

        if response.id != request.id:
            raise exceptions.IdentityError(
                f"response id doesn't match the request one: expected {request.id}, got {response.id}"
            )

        response.related = request


class AbstractAsyncClient(AbstractClient):
    """
    Abstract asynchronous JSON-RPC client.
    """

    @property
    def batch(self):
        """
        Client batch wrapper.
        """

        return AsyncBatch(self)

    @abc.abstractmethod
    async def _request(self, data, **kwargs):
        """
        Makes a JSON-RPC request.

        :param data: request text representation
        :returns: response text representation
        """

    async def send(self, request, **kwargs):
        """
        Sends a JSON-RPC request.

        :param request: request instance
        :param kwargs: additional client request argument
        :returns: response instance
        """

        kwargs = {**self.request_args, **kwargs}

        request_text = self.json_dumper(request, cls=self.json_encoder)
        response_text = await self._request(request_text, **kwargs)

        return self.response_class.from_json(
            self.json_loader(response_text, cls=self.json_decoder), error_cls=self.error_cls
        )

    async def notify(self, method, *args, **kwargs):
        """
        Makes a notification request

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=None,
            method=method,
            params=args or kwargs,
        )
        request_text = self.json_dumper(request, cls=self.json_encoder)
        response_text = await self._request(request_text, **self.request_args)
        if self.strict and response_text:
            raise exceptions.BaseError("unexpected response")

    async def call(self, method, *args, **kwargs):
        """
        Makes JSON-RPC call.

        :param method: method name
        :param args: method positional arguments
        :param kwargs: method named arguments
        :returns: response result
        """

        assert not (args and kwargs), "positional and keyword arguments are mutually exclusive"

        request = self.request_class(
            id=next(self.id_gen()),
            method=method,
            params=args or kwargs,
        )
        response = await self.send(request)
        self._relate(request, response)

        return response.result



"""
Builtin request id generators. Implements several identifier types and generation strategies.
"""

import itertools as it
import random as _random
import string
import uuid as _uuid


def sequential(start=1, step=1):
    """
    Sequential id generator. Returns consecutive values starting from `start` with step `step`.
    """

    yield from it.count(start, step)


def randint(a, b):
    """
    Random integer id generator. Returns random integers between `a` and `b`.
    """

    while True:
        yield _random.randint(a, b)


def random(length=8, chars=string.digits + string.ascii_lowercase):
    """
    Random string id generator. Returns random strings of length `length` using alphabet `chars`.
    """

    while True:
        yield ''.join((_random.choice(chars) for _ in range(length)))


def uuid():
    """
    UUID id generator. Returns random UUIDs.
    """

    while True:
        yield _uuid.uuid4()



import pytest
from pjrpc.common import v20, exceptions


@pytest.mark.parametrize('id, result', [
    (1, None),
    ('1', None),
    (None, None),
    (1, 'result'),
    (1, {'key': 'value'}),
    (1, [1, 2]),
    (1, None),
])
def test_response_serialization(id, result):
    response = v20.Response(id=id, result=result)

    actual_dict = response.to_json()
    expected_dict = {
        'jsonrpc': '2.0',
        'id': id,
        'result': result,
    }

    assert actual_dict == expected_dict


@pytest.mark.parametrize('id, result', [
    (1, None),
    ('1', None),
    (None, None),
    (1, 'result'),
    (1, {'key': 'value'}),
    (1, [1, 2]),
    (1, None),
])
def test_response_deserialization(id, result):
    data = {
        'jsonrpc': '2.0',
        'id': id,
        'result': result,
    }

    response = v20.Response.from_json(data)

    assert response.id == id
    assert response.result == result


def test_response_properties():
    response = v20.Response(id=None, result='result')

    assert response.is_success is True
    assert response.is_error is False

    response = v20.Response(id=None, error='error')

    assert response.is_success is False
    assert response.is_error is True


def test_batch_response_serialization():
    response = v20.BatchResponse(
        v20.Response(id=None, result='result0'),
        v20.Response(id=1, result='result1'),
        v20.Response(id=2, result='result2'),
    )

    actual_dict = response.to_json()
    expected_dict = [
        {
            'jsonrpc': '2.0',
            'id': None,
            'result': 'result0',
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'result1',
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'result': 'result2',
        },
    ]

    assert actual_dict == expected_dict


def test_batch_response_deserialization():
    data = [
        {
            'jsonrpc': '2.0',
            'id': None,
            'result': 'result0',
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'result1',
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'result': 'result2',
        },
    ]

    response = v20.BatchResponse.from_json(data)

    assert response[0].id is None
    assert response[0].result == 'result0'

    assert response[1].id == 1
    assert response[1].result == 'result1'

    assert response[2].id == 2
    assert response[2].result == 'result2'


def test_batch_response_methods():
    request = v20.BatchResponse(
        v20.Response(id=None, result='result0'),
        v20.Response(id=None, result='result0'),
        v20.Response(id=1, result='result1'),
    )
    request.append(v20.Response(id=2, result='result2'))
    request.extend([
        v20.Response(id=3, result='result3'),
        v20.Response(id=4, result='result4'),
    ])


def test_batch_response_errors():
    with pytest.raises(exceptions.IdentityError):
        v20.BatchResponse(
            v20.Response(id=1, result='result1'),
            v20.Response(id=1, result='result1'),
        )



import pytest
from pjrpc.common import v20, exceptions


@pytest.mark.parametrize('id, params', [
    (1, []),
    ('1', []),
    (None, []),
    (1, None),
    (1, []),
    (1, [1, 2]),
    (1, {'a': 1, 'b': 2}),
])
def test_request_serialization(id, params):
    request = v20.Request('method', params, id=id)

    actual_dict = request.to_json()
    expected_dict = {
        'jsonrpc': '2.0',
        'id': id,
        'method': 'method',
    }
    if params:
        expected_dict.update(params=params)

    assert actual_dict == expected_dict


@pytest.mark.parametrize('id, params', [
    (1, []),
    ('1', []),
    (None, []),
    (1, None),
    (1, []),
    (1, [1, 2]),
    (1, {'a': 1, 'b': 2}),
])
def test_request_deserialization(id, params):
    data = {
        'jsonrpc': '2.0',
        'id': id,
        'method': 'method',
    }
    if params:
        data.update(params=params)

    request = v20.Request.from_json(data)

    assert request.id == id
    assert request.method == 'method'
    if params:
        assert request.params == params


def test_request_properties():
    request = v20.Request('method')

    assert request.is_notification is True


def test_batch_request_serialization():
    request = v20.BatchRequest(
        v20.Request('method0', [], id=None),
        v20.Request('method1', [1, 2], id=1),
        v20.Request('method2', {'a': 1, 'b': 2}, id=2)
    )

    actual_dict = request.to_json()
    expected_dict = [
        {
            'jsonrpc': '2.0',
            'id': None,
            'method': 'method0',
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'id': 2,
            'method': 'method2',
            'params': {'a': 1, 'b': 2},
        }
    ]

    assert actual_dict == expected_dict


def test_batch_request_deserialization():
    data = [
        {
            'jsonrpc': '2.0',
            'id': None,
            'method': 'method0',
            'params': [],
        },
        {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method1',
            'params': [1, 2],
        },
        {
            'jsonrpc': '2.0',
            'id': None,
            'method': 'method2',
            'params': {'a': 1, 'b': 2},
        }
    ]

    request = v20.BatchRequest.from_json(data)

    assert request[0].id is None
    assert request[0].method == 'method0'
    assert request[0].params == []

    assert request[1].id == 1
    assert request[1].method == 'method1'
    assert request[1].params == [1, 2]

    assert request[2].id is None
    assert request[2].method == 'method2'
    assert request[2].params == {'a': 1, 'b': 2}


def test_batch_request_methods():
    request = v20.BatchRequest(
        v20.Request('method1', [1], id=None),
        v20.Request('method2', [1], id=None),
        v20.Request('method3', [1], id=1),
    )
    request.append(v20.Request('method4', [2], id=2))
    request.extend([
        v20.Request('method5', [3], id=3),
        v20.Request('method6', [4], id=4),
    ])


def test_batch_request_errors():
    with pytest.raises(exceptions.IdentityError):
        v20.BatchRequest(
            v20.Request(id=1, method='method'),
            v20.Request(id=1, method='method'),
        )



"""
Standard python http server JSON-RPC integration.
"""

import http.server

import pjrpc


class JsonRpcHandler(http.server.BaseHTTPRequestHandler):
    """
    JSON-RPC handler.
    """

    def do_POST(self):
        """
        Handles JSON-RPC request.
        """

        content_type = self.headers.get('Content-Type')
        if content_type != 'application/json':
            self.send_response(http.HTTPStatus.UNSUPPORTED_MEDIA_TYPE)
            return

        try:
            content_length = int(self.headers.get('Content-Length', -1))
            request_text = self.rfile.read(content_length).decode()
        except UnicodeDecodeError:
            self.send_response(http.HTTPStatus.BAD_REQUEST)
            return

        response_text = self.server.dispatcher.dispatch(request_text, context=self)
        if response_text is None:
            self.send_response(http.HTTPStatus.OK)
        else:
            self.send_response(http.HTTPStatus.OK)
            self.send_header("Content-type", "application/json")
            self.end_headers()

            self.wfile.write(response_text.encode())


class JsonRpcServer(http.server.HTTPServer):
    """
    :py:class:`http.server.HTTPServer` based JSON-RPC server.

    :param path: JSON-RPC handler base path
    :param kwargs: arguments to be passed to the dispatcher :py:class:`pjrpc.server.Dispatcher`
    """

    def __init__(self, server_address, RequestHandlerClass=JsonRpcHandler, bind_and_activate=True, **kwargs):
        super().__init__(server_address, RequestHandlerClass, bind_and_activate)
        self._dispatcher = pjrpc.server.Dispatcher(**kwargs)

    @property
    def dispatcher(self):
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher



import json
import yarl
from aioresponses import aioresponses

import pjrpc
from pjrpc.client.backend import aiohttp as pjrpc_cli


async def test_call():
    test_url = 'http://test.com/api'
    with aioresponses() as m:
        m.post(test_url, repeat=True, status=200, payload={
            'jsonrpc': '2.0',
            'id': 1,
            'result': 'result',
        })

        client = pjrpc_cli.Client(test_url)

        response = await client.send(pjrpc.Request('method', (1, 2), id=1))

        assert response.id == 1
        assert response.result == 'result'

        assert len(m.requests) == 1
        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[0].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2]
        }

        result = await client.call('method', arg1=1, arg2=2)
        assert result == 'result'

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[1].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': {'arg1': 1, 'arg2': 2}
        }

        result = await client('method', 1, 2)
        assert result == 'result'

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[2].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2]
        }

        result = await client.proxy.method(1, 2)
        assert result == 'result'

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[3].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': 1,
            'method': 'method',
            'params': [1, 2]
        }


async def test_notify():
    test_url = 'http://test.com/api'
    with aioresponses() as m:
        m.post(test_url, repeat=True, status=200, body='')

        client = pjrpc_cli.Client(test_url)

        await client.notify('method', 1, 2)

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[0].kwargs['data']) == {
            'jsonrpc': '2.0',
            'id': None,
            'method': 'method',
            'params': [1, 2]
        }


async def test_batch():
    test_url = 'http://test.com/api'
    with aioresponses() as m:
        m.post(test_url, repeat=True, status=200, payload=[
            {
                'jsonrpc': '2.0',
                'id': 1,
                'result': 'result1',
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'result': 2,
            }
        ])

        client = pjrpc_cli.Client(test_url)

        result = await client.batch[
            ('method1', 1, 2),
            ('method2', 2, 3),
        ]
        assert result == ('result1', 2)

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[0].kwargs['data']) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2]
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3]
            }
        ]

        result = await client.batch('method1', 1, 2)('method2', 2, 3).call()
        assert result == ('result1', 2)

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[1].kwargs['data']) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2]
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3]
            }
        ]

        result = await client.batch.proxy.method1(1, 2).method2(2, 3)()
        assert result == ('result1', 2)

        requests = m.requests[('POST', yarl.URL(test_url))]
        assert json.loads(requests[2].kwargs['data']) == [
            {
                'jsonrpc': '2.0',
                'id': 1,
                'method': 'method1',
                'params': [1, 2]
            },
            {
                'jsonrpc': '2.0',
                'id': 2,
                'method': 'method2',
                'params': [2, 3]
            }
        ]



import logging
import functools as ft

import aio_pika

import pjrpc

logger = logging.getLogger(__package__)


class Executor:
    """
    `aio_pika <https://aio-pika.readthedocs.io/en/latest/>`_ based JSON-RPC server.

    :param broker_url: broker connection url
    :param queue_name: requests queue name
    :param prefetch_count: worker prefetch count
    :param workers_count: number of workers
    :param kwargs: dispatcher additional arguments
    """

    class Worker:
        """
        Executor worker.
        """

        def __init__(self, connection, queue_name, handler, prefetch_count=0):
            self._connection = connection
            self._queue_name = queue_name
            self._handler = handler
            self._prefetch_count = prefetch_count

            self._channel = None
            self._queue = None
            self._consumer_tag = None

        async def start(self):
            self._channel = await self._connection.channel()
            await self._channel.set_qos(prefetch_count=self._prefetch_count)

            self._queue = await self._channel.declare_queue(self._queue_name, passive=True)
            self._consumer_tag = await self._queue.consume(ft.partial(self._handler, channel=self._channel))

        async def stop(self):
            if self._consumer_tag:
                await self._queue.cancel(self._consumer_tag)
            if self._channel:
                await self._channel.close()

    def __init__(self, broker_url, queue_name, prefetch_count=0, workers_count=1, **kwargs):
        self._broker_url = broker_url
        self._queue_name = queue_name
        self._prefetch_count = prefetch_count

        self._connection = aio_pika.connection.Connection(broker_url)
        self._workers = [
            Executor.Worker(
                self._connection, queue_name, self._rpc_handle, prefetch_count,
            ) for _ in range(workers_count)
        ]

        self._dispatcher = pjrpc.server.Dispatcher(**kwargs)

    @property
    def dispatcher(self):
        """
        JSON-RPC method dispatcher.
        """

        return self._dispatcher

    async def start(self, queue_args=None):
        """
        Starts executor.

        :param queue_args: queue arguments
        """

        await self._connection.connect()
        async with self._connection.channel() as channel:
            await channel.declare_queue(self._queue_name, **(queue_args or {}))

        for worker in self._workers:
            await worker.start()

    async def shutdown(self):
        """
        Stops executor.
        """

        for worker in self._workers:
            await worker.stop()

        await self._connection.close()

    async def _rpc_handle(self, message, channel):
        """
        Handles JSON-RPC request.

        :param channel: broker channel
        :param message: incoming message
        """

        try:
            reply_to = message.reply_to
            response_text = self._dispatcher.dispatch(message.body, context=message)

            if response_text is not None:
                if reply_to is None:
                    logger.warning("property 'reply_to' is missing")
                else:
                    await channel.default_exchange.publish(
                        aio_pika.Message(
                            body=response_text.encode(),
                            reply_to=reply_to,
                            correlation_id=message.correlation_id,
                            content_type='application/json',
                        ),
                        routing_key=reply_to
                    )

            message.ack()

        except Exception as e:
            logger.exception("jsonrpc request handling error: %s", e)
